@@ -1,8 +1,8 @@
 # Module: MachineLearning
 # Author: Adrian Antico <adrianantico@gmail.com>
 # License: MIT
-# Release: retrofit 0.1.4
-# Last modified : 2021-09-15
+# Release: retrofit 0.1.5
+# Last modified : 2021-09-20
 
 def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsList=None, TargetColumnName=None, NumericColumnNames=None, CategoricalColumnNames=None, TextColumnNames=None, WeightColumnName=None, Threads=-1, Processing='catboost', InputFrame='datatable'):
     """
@@ -27,6 +27,7 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
     InputFrame:             'datatable', 'polars', or 'pandas' If you input Frame is 'pandas', it will be converted to a datatable Frame for generating the new columns
 
     # ML0_GetModelData Example:
+    import pkg_resources
     import timeit
     import datatable as dt
     from datatable import sort, f, by
@@ -35,7 +36,8 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
     from retrofit import MachineLearning as ml
 
     # Load some data
-    data = dt.fread("C:/Users/Bizon/Documents/GitHub/BenchmarkData.csv")
+    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
+    data = dt.fread(FilePath)
         
     # Create partitioned data sets
     DataSets = fe.FE2_AutoDataParition(
@@ -96,6 +98,20 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
     InputFrame='datatable'
     """
     
+    # For making copies of lists so originals aren't modified
+    import copy
+    
+    # Import datatable methods
+    if InputFrame.lower() == 'datatable':
+      import datatable as dt
+      from datatable import sort, f, by, ifelse, join
+
+    # Import polars methods
+    if InputFrame.lower() == 'polars':
+      import polars as pl
+      from polars import col
+      from polars.lazy import col
+
     # ArgsList Collection
     if not ArgsList is None:
       TargetColumnName = ArgsList['TargetColumnName']
@@ -114,20 +130,41 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
         WeightColumnName=WeightColumnName,
         Threads=Threads,
         Processing=Processing)
-
-    # For making copies of lists so originals aren't modified
-    import copy
     
-    # Import datatable methods
-    if InputFrame.lower() == 'datatable':
-      import datatable as dt
-      from datatable import sort, f, by, ifelse
-
-    # Import polars methods
-    if InputFrame.lower() == 'polars':
-      import polars as pl
-      from polars import col
-      from polars.lazy import col
+    # Target Variable Conversion for Multiclass
+    if TrainData.types[TrainData.colindex(TargetColumnName)] in [dt.Type.str32, dt.Type.str64, dt.Type.is_string]:
+      def MultiClassTargetToInt(TrainData=None, ValidationData=None, TestData=None, TargetColumnName=None, ArgsList=None):
+        import numpy as np
+        import datatable as dt
+        from datatable import f, by, rbind, join
+        temp = TrainData[:, TargetColumnName, by(TargetColumnName)]
+        if not ValidationData is None:
+          temp2 = ValidationData[:, TargetColumnName, by(TargetColumnName)]
+          temp.rbind(temp2)
+          del temp2
+        if not TestData is None:
+          temp3 = TestData[:, TargetColumnName, by(TargetColumnName)]
+          temp.rbind(temp3)
+          del temp3
+        temp = temp[:, TargetColumnName, by(TargetColumnName)]
+        del temp[:, temp.names[1]]
+        temp = temp.sort(TargetColumnName)
+        temp[f"Predict_{TargetColumnName}"] = np.arange(0,temp.shape[0], 1)
+        temp.key = TargetColumnName
+        ArgsList['MultiClass'] = temp
+        TrainData = TrainData[:, :, join(temp)]
+        del TrainData[:, TargetColumnName]
+        TrainData.names = {f"Predict_{TargetColumnName}": TargetColumnName}
+        if not ValidationData is None:
+          ValidationData = ValidationData[:, :, join(temp)]
+          del ValidationData[:, TargetColumnName]
+          ValidationData.names = {f"Predict_{TargetColumnName}": TargetColumnName}
+        if not TestData is None:
+          TestData = TestData[:, :, join(temp)]
+          del TestData[:, TargetColumnName]
+          TestData.names = {f"Predict_{TargetColumnName}": TargetColumnName}
+        temp.names = {TargetColumnName: 'Old'}
+        return dict(TrainData = TrainData, ValidationData = ValidationData, TestData = TestData, ArgsList = ArgsList)
 
     # Convert to datatable
     if InputFrame.lower() == 'pandas' and Processing.lower() == 'datatable': 
@@ -189,7 +226,15 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
       if not TestData is None:
         test = TestData[:, SD].to_pandas()
 
-      # label
+      # Categorical target check
+      if TrainData.types[TrainData.colindex(TargetColumnName)] in [dt.Type.str32, dt.Type.str64, dt.Type.is_string]:
+        Output = MultiClassTargetToInt(TrainData=TrainData, ValidationData=ValidationData, TestData=TestData, TargetColumnName=TargetColumnName, ArgsList=ArgsList)
+        TrainData = Output['TrainData']
+        ValidationData = Output['ValidationData']
+        TestData = Output['TestData']
+        ArgsList = Output['ArgsList']
+
+      # Labels
       trainlabel = TrainData[:, TargetColumnName].to_pandas()
       if not ValidationData is None:
         validationlabel = ValidationData[:, TargetColumnName].to_pandas()
@@ -252,7 +297,7 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
         trainweightdata = None
         validationweightdata = None
         testweightdata = None
-        
+
       # data
       train = TrainData[:, SD].to_pandas()
       if not ValidationData is None:
@@ -260,7 +305,15 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
       if not TestData is None:
         test = TestData[:, SD].to_pandas()
 
-      # label
+      # Categorical target check
+      if TrainData.types[TrainData.colindex(TargetColumnName)] in [dt.Type.str32, dt.Type.str64, dt.Type.is_string]:
+        Output = MultiClassTargetToInt(TrainData=TrainData, ValidationData=ValidationData, TestData=TestData, TargetColumnName=TargetColumnName, ArgsList=ArgsList)
+        TrainData = Output['TrainData']
+        ValidationData = Output['ValidationData']
+        TestData = Output['TestData']
+        ArgsList = Output['ArgsList']
+
+      # Target label
       trainlabel = TrainData[:, TargetColumnName].to_pandas()
       if not ValidationData is None:
         validationlabel = ValidationData[:, TargetColumnName].to_pandas()
@@ -319,6 +372,14 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
       if not TestData is None:
         test = TestData[:, SD].to_pandas()
 
+      # Categorical target check
+      if TrainData.types[TrainData.colindex(TargetColumnName)] in [dt.Type.str32, dt.Type.str64, dt.Type.is_string]:
+        Output = MultiClassTargetToInt(TrainData=TrainData, ValidationData=ValidationData, TestData=TestData, TargetColumnName=TargetColumnName, ArgsList=ArgsList)
+        TrainData = Output['TrainData']
+        ValidationData = Output['ValidationData']
+        TestData = Output['TestData']
+        ArgsList = Output['ArgsList']
+
       # label
       trainlabel = TrainData[:, TargetColumnName].to_pandas()
       if not ValidationData is None:
@@ -350,7 +411,7 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
       return dict(train_data=train_data, validation_data=validation_data, test_data=test_data, ArgsList=ArgsList)
 
 
-def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=None):
+def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None):
     """
     # Goal
     Return an ArgsList appropriate for the algorithm selection, target type, and training method
@@ -359,9 +420,9 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
     Algorithms:       Choose from CatBoost, XGBoost, LightGBM, Ftrl
     TargetType:       Choose from 'regression', 'classification', 'multiclass'
     TrainMethod:      Choose from 'train', 'gridtune'
-    GetModelDataArgs: Args passed in from ML0_GetModelData()
     
     # ML0_Parameters Example
+    import pkg_resources
     import timeit
     import datatable as dt
     from datatable import sort, f, by
@@ -370,7 +431,8 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
     from retrofit import MachineLearning as ml
 
     # Load some data
-    data = dt.fread("C:/Users/Bizon/Documents/GitHub/BenchmarkData.csv")
+    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
+    data = dt.fread(FilePath)
         
     # Create partitioned data sets
     DataSets = fe.FE2_AutoDataParition(
@@ -412,8 +474,7 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
     ModelArgs = ml.ML0_Parameters(
       Algorithms='catboost', 
       TargetType='regression', 
-      TrainMethod='Train', 
-      Model=None)
+      TrainMethod='Train')
 
     # QA
     Algorithms='catboost'
@@ -423,6 +484,14 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
     Algo = 'catboost'
     """
     
+    # Args Check
+    if Algorithms is None:
+      raise Exception('Algorithms cannot be None')
+    if TargetType is None:
+      raise Exception('TargetType cannot be None')
+    if TrainMethod is None:
+      raise Exception('TrainMethod cannot be None')  
+    
     # Ensure Algorithms is a list
     if not isinstance(Algorithms, list):
       Algorithms = [Algorithms]
@@ -453,13 +522,13 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
         # TargetType Parameters
         ###############################
         if ArgsList.get('TargetType').lower() == 'classification':
+          AlgoArgs['loss_function'] = 'Logloss'
+          AlgoArgs['eval_metric'] = 'Logloss'
           AlgoArgs['auto_class_weights'] = 'Balanced'
-          AlgoArgs['loss_function'] = 'CrossEntropy'
-          AlgoArgs['eval_metric'] = 'CrossEntropy'
         elif ArgsList.get('TargetType').lower() == 'multiclass':
           AlgoArgs['classes_count'] = 3
-          AlgoArgs['loss_function'] = 'MCC'
-          AlgoArgs['eval_metric'] = 'MCC'
+          AlgoArgs['loss_function'] = 'MultiClassOneVsAll'
+          AlgoArgs['eval_metric'] = 'MultiClassOneVsAll'
         elif ArgsList.get('TargetType').lower() == 'regression':
           AlgoArgs['loss_function'] = 'RMSE'
           AlgoArgs['eval_metric'] = 'RMSE'
@@ -596,7 +665,7 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
         if AlgoArgs['tree_method'] == 'gpu_hist':
           AlgoArgs['sampling_method'] = 'uniform'
 
-        # Classification
+        # Target Dependent Args
         if ArgsList.get('TargetType').lower() == 'classification':
           AlgoArgs['objective'] = 'binary:logistic'
           AlgoArgs['eval_metric'] = 'auc'
@@ -621,6 +690,17 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
         import lightgbm as lgbm
         AlgoArgs = dict()
         
+        # Target Dependent Args
+        if ArgsList.get('TargetType').lower() == 'classification':
+          AlgoArgs['objective'] = 'binary'
+          AlgoArgs['metric'] = 'auc'
+        elif ArgsList.get('TargetType').lower() == 'regression':
+          AlgoArgs['objective'] = 'regression'
+          AlgoArgs['metric'] = 'rmse'
+        elif ArgsList.get('TargetType').lower() == 'multiclass':
+          AlgoArgs['objective'] = 'multiclassova'
+          AlgoArgs['metric'] = 'multi_logloss'
+
         # Tuning Args
         if TrainMethod.lower() == 'train':
           AlgoArgs['num_iterations'] = 1000
@@ -644,8 +724,6 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None, Model=Non
         # Args
         AlgoArgs['task'] = 'train'
         AlgoArgs['device_type'] = 'CPU'
-        AlgoArgs['objective'] = 'regression'
-        AlgoArgs['metric'] = 'rmse'
         AlgoArgs['boosting'] = 'gbdt'
         AlgoArgs['lambda_l1'] = 0.0
         AlgoArgs['lambda_l2'] = 0.0
@@ -801,6 +879,7 @@ class RetroFit:
     ####################################
     
     # Setup Environment
+    import pkg_resources
     import timeit
     import datatable as dt
     from datatable import sort, f, by
@@ -810,9 +889,8 @@ class RetroFit:
     
     # Load some data
     # BechmarkData.csv is located is the tests folder
-    Path = "./BenchmarkData.csv"
-    data = dt.fread(Path)
-    data = dt.fread("C:/Users/Bizon/Documents/GitHub/BenchmarkData.csv")
+    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
+    data = dt.fread(FilePath)
 
     # Create partitioned data sets
     Data = fe.FE2_AutoDataParition(
@@ -881,6 +959,7 @@ class RetroFit:
     ####################################
     
     # Setup Environment
+    import pkg_resources
     import timeit
     import datatable as dt
     from datatable import sort, f, by
@@ -889,7 +968,8 @@ class RetroFit:
     from retrofit import MachineLearning as ml
     
     # Load some data
-    data = dt.fread("C:/Users/Bizon/Documents/GitHub/BenchmarkData.csv")
+    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
+    data = dt.fread(FilePath)
     
     # Create partitioned data sets
     DataFrames = fe.FE2_AutoDataParition(
@@ -957,6 +1037,7 @@ class RetroFit:
     ####################################
     
     # Setup Environment
+    import pkg_resources
     import timeit
     import datatable as dt
     from datatable import sort, f, by
@@ -965,7 +1046,8 @@ class RetroFit:
     from retrofit import MachineLearning as ml
     
     # Load some data
-    data = dt.fread("C:/Users/Bizon/Documents/GitHub/BenchmarkData.csv")
+    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
+    data = dt.fread(FilePath)
     
     # Create partitioned data sets
     DataFrames = fe.FE2_AutoDataParition(
@@ -1036,6 +1118,7 @@ class RetroFit:
     ####################################
     
     # Setup Environment
+    import pkg_resources
     import timeit
     import datatable as dt
     from datatable import sort, f, by
@@ -1044,7 +1127,8 @@ class RetroFit:
     from retrofit import MachineLearning as ml
     
     # Load some data
-    data = dt.fread("C:/Users/Bizon/Documents/GitHub/BenchmarkData.csv")
+    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
+    data = dt.fread(FilePath)
     
     # Dummify
     Output = fe.FE1_DummyVariables(
@@ -1140,7 +1224,6 @@ class RetroFit:
       self.InterpretationListNames = []
       self.CompareModelsList = dict()
       self.CompareModelsListNames = []
-
     
     #################################################
     #################################################
@@ -1160,12 +1243,11 @@ class RetroFit:
       
       # Check
       if len(self.ModelArgs) == 0:
-        return print("ModelArgs is empty")
+        raise Exception('self.ModelArgs is empty')
 
       # Which Algo
       if not Algorithm is None:
         TempArgs = self.ModelArgs[Algorithm]
-        #TempArgs = ModelArgs[Algorithm]
       else:
         TempArgs = self.ModelArgs[[*self.ModelArgs][0]]
 
@@ -1175,7 +1257,7 @@ class RetroFit:
       if TempArgs.get('Algorithms').lower() == 'ftrl':
 
         # Setup Environment
-        import datatable
+        import datatable as dt
         from datatable import f
         from datatable.models import Ftrl
 
@@ -1212,7 +1294,11 @@ class RetroFit:
         # Initialize model
         if TempArgs.get('TargetType').lower() == 'regression':
           Model = CatBoostRegressor(**TempArgs.get('AlgoArgs'))
-        else:
+        elif TempArgs.get('TargetType').lower() == 'classification':
+          Model = CatBoostClassifier(**TempArgs.get('AlgoArgs'))
+        elif TempArgs.get('TargetType').lower() == 'multiclass':
+          self.ModelArgs.get('CatBoost').get('AlgoArgs')['classes_count'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
+          TempArgs.get('AlgoArgs')['classes_count'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
           Model = CatBoostClassifier(**TempArgs.get('AlgoArgs'))
         
         # Store Model
@@ -1236,10 +1322,15 @@ class RetroFit:
         TrainData = self.DataSets.get('train_data')
         ValidationData = self.DataSets.get('validation_data')
         TestData = self.DataSets.get('test_data')
-        
+
+        # Update args for multiclass
+        if TempArgs.get('TargetType').lower() == 'multiclass':
+          self.ModelArgs.get('XGBoost').get('AlgoArgs')['num_class'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
+          TempArgs.get('AlgoArgs')['num_class'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
+
         # Initialize model
         Model = xgb.XGBModel(**TempArgs.get('AlgoArgs'))
-          
+
         # Store Model
         self.ModelList[f"XGBoost{str(len(self.ModelList) + 1)}"] = Model
         self.ModelListNames.append(f"XGBoost{str(len(self.ModelList))}")
@@ -1262,18 +1353,28 @@ class RetroFit:
         ValidationData = self.DataSets.get('validation_data')
         TestData = self.DataSets.get('test_data')
 
+        # Create temp args
+        import copy
+        temp_args = copy.deepcopy(TempArgs)
+
+        # Update args for multiclass
+        if TempArgs.get('TargetType').lower() == 'multiclass':
+          self.ModelArgs.get('LightGBM').get('AlgoArgs')['num_class'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
+          TempArgs.get('AlgoArgs')['num_class'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
+          temp_args.get('AlgoArgs')['num_class'] = self.DataSets.get('ArgsList')['MultiClass'].shape[0]
+
+        # Create modified args
+        del temp_args['AlgoArgs']['num_iterations']
+        del temp_args['AlgoArgs']['early_stopping_round']
+
         # Initialize model
-        Model = LGBMModel(**TempArgs.get('AlgoArgs'))
-        
+        Model = LGBMModel(**temp_args.get('AlgoArgs'))
+
         # Store Model
         self.ModelList[f"LightGBM{str(len(self.ModelList) + 1)}"] = Model
         self.ModelListNames.append(f"LightGBM{str(len(self.ModelList))}")
 
         # Initialize model
-        import copy
-        temp_args = copy.deepcopy(TempArgs)
-        del temp_args['AlgoArgs']['num_iterations']
-        del temp_args['AlgoArgs']['early_stopping_round']
         self.FitList[f"LightGBM{str(len(self.FitList) + 1)}"] = lgbm.train(params=temp_args.get('AlgoArgs'), train_set=TrainData, valid_sets=[ValidationData, TestData], num_boost_round=TempArgs.get('AlgoArgs').get('num_iterations'), early_stopping_rounds=TempArgs.get('AlgoArgs').get('early_stopping_round'))
         self.FitListNames.append(f"LightGBM{str(len(self.FitList))}")
 
@@ -1286,7 +1387,7 @@ class RetroFit:
 
       # Check
       if len(self.ModelList) == 0:
-        return print("No models exist")
+        raise Exception('No models found in self.ModelList')
 
       # Which Algo
       if not Algorithm is None:
@@ -1295,19 +1396,22 @@ class RetroFit:
         TempArgs = self.ModelArgs[[*self.ModelArgs][0]]
 
       # Setup Environment
-      import datatable
+      import datatable as dt
       from datatable.models import Ftrl
 
       #################################################
       # Ftrl Method
       #################################################
       if TempArgs['Algorithms'].lower() == 'ftrl':
+
+        # Setup Environment
+        from datatable import f
         
         # Extract model
         if not ModelName is None:
           Model = self.ModelList.get(ModelName)
         else:
-          Model = self.ModelList.get(f"Ftrl_{str(len(self.FitList))}")
+          Model = self.ModelList.get(f"Ftrl{str(len(self.FitList))}")
 
         # Grab scoring data
         TargetColumnName = self.DataSets.get('ArgsList')['TargetColumnName']
@@ -1320,14 +1424,21 @@ class RetroFit:
         if TargetColumnName in score_data.names:
           TargetData = score_data[:, f[TargetColumnName]]
           score_data = score_data[:, f[:].remove(f[TargetColumnName])]
-          if NewData is None:
-            return Model.predict(score_data)
 
         # Score Model and append data set name to scoring data
-        score_data.cbind(Model.predict(score_data))
-        
-        # Update prediction column name
-        score_data.names = {TargetColumnName: f"Predict_{TargetColumnName}"}
+        if self.ModelArgs.get('Ftrl').get('TargetType').lower() == 'regression':
+          score_data.cbind(Model.predict(score_data))
+          score_data.names = {TargetColumnName: f"Predict_{TargetColumnName}"}
+        elif self.ModelArgs.get('Ftrl').get('TargetType').lower() == 'classification':
+          score_data.cbind(Model.predict(score_data))
+          score_data.names = {'1.0': 'p1'}
+          score_data.names = {'0.0': 'p0'}
+        elif self.ModelArgs.get('Ftrl').get('TargetType').lower() == 'multiclass':
+          score_data.cbind(Model.predict(score_data))
+
+        # Return preds
+        if not NewData is None:
+          return ScoreData
         
         # cbind Target column back to score_data
         score_data.cbind(TargetData)
@@ -1345,7 +1456,7 @@ class RetroFit:
         if not ModelName is None:
           Model = self.ModelList.get(ModelName)
         else:
-          Model = self.ModelList.get(f"CatBoost_{str(len(self.FitList))}")
+          Model = self.ModelList.get(f"CatBoost{str(len(self.FitList))}")
 
         # Grab dataframe data
         TargetColumnName = self.DataSets.get('ArgsList')['TargetColumnName']
@@ -1361,20 +1472,26 @@ class RetroFit:
           pred_data = NewData
 
         # Generate preds and add to datatable frame
-        if NewData is None:
-          if TempArgs.get('TargetType').lower() == 'regression':
-            ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(pred_data, prediction_type = 'RawFormulaVal')
-          elif TempArgs.get('TargetType').lower() == 'classification':
-            ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(pred_data, prediction_type = 'Probability')
-          elif TempArgs.get('TargetType').lower() == 'multiclass':
-            ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(pred_data, prediction_type = 'Class')
-        else:
-          if TempArgs.get('TargetType').lower() == 'regression':
-            return Model.predict(pred_data, prediction_type = 'RawFormulaVal')
-          elif TempArgs.get('TargetType').lower() == 'classification':
-            return Model.predict(pred_data, prediction_type = 'Probability')
-          elif TempArgs.get('TargetType').lower() == 'multiclass':
-            return Model.predict(pred_data, prediction_type = 'Class')
+        if TempArgs.get('TargetType').lower() == 'regression':
+          ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(pred_data, prediction_type = 'RawFormulaVal')
+        elif TempArgs.get('TargetType').lower() == 'classification':
+          temp = Model.predict(pred_data, prediction_type = 'Probability')
+          ScoreData['p0'] = temp[:,0]
+          ScoreData['p1'] = temp[:,1]
+        elif TempArgs.get('TargetType').lower() == 'multiclass':
+          ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(pred_data, prediction_type = 'Class')
+          if not self.DataSets.get('ArgsList')['MultiClass'] is None:
+            from datatable import join
+            temp = self.DataSets.get('ArgsList')['MultiClass']
+            temp.key = f"Predict_{TargetColumnName}"
+            ScoreData = ScoreData[:, :, join(temp)]
+            temp_x = f"Predict_{TargetColumnName}"
+            del ScoreData[:, temp_x]
+            ScoreData.names = {'Old': f"Predict_{TargetColumnName}"}
+
+        # Return preds
+        if not NewData is None:
+          return ScoreData
 
         # Store data and update names
         self.DataSets[f"Scored_{DataName}_{Algorithm}_{len(self.FitList)}"] = ScoreData
@@ -1387,12 +1504,13 @@ class RetroFit:
 
         # Environment
         import xgboost as xgb
+        from datatable import f
 
         # Extract Model
         if not ModelName is None:
           Model = self.FitList.get(ModelName)
         else:
-          Model = self.FitList.get(f"XGBoost_{str(len(self.FitList))}")
+          Model = self.FitList.get(f"XGBoost{str(len(self.FitList))}")
 
         # Grab dataframe data
         TargetColumnName = self.DataSets.get('ArgsList')['TargetColumnName']
@@ -1405,11 +1523,11 @@ class RetroFit:
           elif DataName == 'train_data':
             ScoreData = self.DataFrames.get('TrainData')
         else:
-          score_data = NewData
+          ScoreData = NewData
           pred_data = self.DataSets[DataName]
 
         # Generate preds and add to datatable frame
-        if NewData is None:
+        if TempArgs.get('TargetType').lower() != 'multiclass':
           ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(
             data = pred_data, 
             output_margin=False, 
@@ -1421,8 +1539,14 @@ class RetroFit:
             training=False, 
             iteration_range=(0, self.FitList[f"XGBoost{str(len(self.FitList))}"].best_iteration), 
             strict_shape=False)
+          
+          # Classification
+          if TempArgs.get('TargetType').lower() == 'classification':
+            ScoreData.names = {f"Predict_{TargetColumnName}": "p1"}
+            ScoreData = ScoreData[:, f[:].extend({'p0': 1 - f['p1']})]
+          
         else:
-          return Model.predict(
+          preds = dt.Frame(Model.predict(
             data = pred_data, 
             output_margin=False, 
             pred_leaf=False, 
@@ -1432,8 +1556,24 @@ class RetroFit:
             validate_features=True, 
             training=False, 
             iteration_range=(0, self.FitList[f"XGBoost{str(len(self.FitList))}"].best_iteration), 
-            strict_shape=False)
-        
+            strict_shape=False))
+
+          # MultiClass Case
+          if not self.DataSets.get('ArgsList')['MultiClass'] is None:
+            from datatable import cbind
+            temp = self.DataSets.get('ArgsList')['MultiClass']
+            counter = 0
+            for val in temp['Old'].to_list()[0]:
+              preds.names = {f"C{counter}": val}
+              counter += 1
+  
+            # Combine ScoreData and preds
+            ScoreData.cbind(preds)
+
+        # Return preds
+        if not NewData is None:
+          return ScoreData
+
         # Store data and update names
         self.DataSets[f"Scored_{DataName}_{Algorithm}_{len(self.FitList)}"] = ScoreData
         self.DataSetsNames.append(f"Scored_{DataName}_{Algorithm}_{len(self.FitList)}")
@@ -1445,6 +1585,7 @@ class RetroFit:
         
         # Environment
         import lightgbm as lgbm
+        from datatable import f
         
         # Extract Model
         if not ModelName is None:
@@ -1461,17 +1602,102 @@ class RetroFit:
             ScoreData = self.DataFrames.get('ValidationData')
           elif DataName == 'train_data':
             ScoreData = self.DataFrames.get('TrainData')
-          ScoreData = ScoreData[:, self.DataSets.get('ArgsList').get('NumericColumnNames')]
         else:
           ScoreData = NewData
-          ScoreData = ScoreData[:, self.DataSets.get('ArgsList').get('NumericColumnNames')]
 
-        # Generate preds and add to datatable frame
-        if NewData is None:
-          ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(data = ScoreData)
-        else:
-          return Model.predict(data = ScoreData)
+        # Subset score data columns
+        ScoreData = ScoreData[:, self.DataSets.get('ArgsList').get('NumericColumnNames')]
         
+        # Regression and Classification
+        if TempArgs.get('TargetType').lower() != 'multiclass':
+          ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(data = ScoreData)
+          
+          # Non regression cases
+          if TempArgs.get('TargetType').lower() == 'classification':
+            ScoreData.names = {f"Predict_{TargetColumnName}": "p1"}
+            ScoreData = ScoreData[:, f[:].extend({'p0': 1 - f['p1']})]
+
+        # MultiClass
+        else:
+          preds = dt.Frame(Model.predict(data = ScoreData))
+          if not self.DataSets.get('ArgsList')['MultiClass'] is None:
+            from datatable import cbind
+            temp = self.DataSets.get('ArgsList')['MultiClass']
+            counter = 0
+            for val in temp['Old'].to_list()[0]:
+              preds.names = {f"C{counter}": val}
+              counter += 1
+
+            # Combine ScoreData and preds
+            ScoreData.cbind(preds)
+
+        # Return preds
+        if not NewData is None:
+          return ScoreData
+
         # Store data and update names
         self.DataSets[f"Scored_{DataName}_{Algorithm}_{len(self.FitList)}"] = ScoreData
         self.DataSetsNames.append(f"Scored_{DataName}_{Algorithm}_{len(self.FitList)}")
+    
+    #################################################
+    #################################################
+    # Function: Evaluation
+    #################################################
+    #################################################
+    
+    # regression metrics helper
+    def _regression_metrics(self, _FitName = None, y_true = None, y_pred = None):
+      import datatable as dt
+      from datetime import datetime
+      from sklearn.metrics import explained_variance_score, max_error, mean_absolute_error, mean_squared_error, mean_squared_log_error, mean_absolute_percentage_error, median_absolute_error, r2_score
+      Metrics = dt.Frame([self.FitList[_FitName]])
+      Metrics.names = {'C0': 'ModelName'}
+      Metrics['FeatureSet'] = None
+      Metrics['CreateTime'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+      Metrics['explained_variance_score'] = explained_variance_score(y_true, y_pred)
+      Metrics['r2_score'] = r2_score(y_true, y_pred)
+      Metrics['mean_absolute_percentage_error'] = mean_absolute_percentage_error(y_true, y_pred)
+      Metrics['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)
+      Metrics['median_absolute_error'] = median_absolute_error(y_true, y_pred)
+      Metrics['mean_squared_error'] = mean_squared_error(y_true, y_pred)
+      Metrics['mean_squared_log_error'] = mean_squared_log_error(y_true, y_pred)
+      Metrics['max_error'] = max_error(y_true, y_pred)
+      return(Metrics)
+    
+    # classification metrics helper
+    def _classification_metrics(self, _FitName = None, y_true = None, y_pred = None):
+      import datatable
+      from datetime import datetime
+      from sklearn.metrics import accuracy_score, balanced_accuracy_score
+      Metrics = dt.Frame([self.FitList[_FitName]])
+      Metrics.names = {'C0': 'ModelName'}
+      Metrics['CreateTime'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+      Metrics['accuracy_score'] = accuracy_score(y_true, y_pred)
+      
+      y_true.to_list()
+      
+      Metrics['balanced_accuracy_score'] = balanced_accuracy_score(y_true, y_pred)
+      Metrics['mean_absolute_percentage_error'] = mean_absolute_percentage_error(y_true, y_pred)
+      Metrics['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)
+      Metrics['median_absolute_error'] = median_absolute_error(y_true, y_pred)
+      Metrics['mean_squared_error'] = mean_squared_error(y_true, y_pred)
+      Metrics['mean_squared_log_error'] = mean_squared_log_error(y_true, y_pred)
+      Metrics['max_error'] = max_error(y_true, y_pred)
+      return(Metrics)
+    
+    # Evaluation Attribute Update
+    def ML1_Single_Evaluate(self, FitName=None, TargetType=None, ScoredDataName=None, ByVariables=None):
+      
+      # Get Data
+      temp = self.DataSets.get(ScoredDataName)
+        
+      # Generate metrics
+      if TargetType == 'regression':
+        Metrics = _regression_metrics(self, _FitName = FitName, y_true = temp[TargetColumnName], y_pred = temp[f"Predict_{TargetColumnName}"])
+      elif TargetType == 'classification':
+        Metrics = _classification_metrics(self, _FitName = FitName, y_true = temp[TargetColumnName], y_pred = temp[f"Predict_{TargetColumnName}"])
+      elif TargetType == 'multiclass':
+        Metrics = _multiclass_metrics(self, _FitName = FitName, y_true = temp[TargetColumnName], y_pred = temp[f"Predict_{TargetColumnName}"])
+      
+      # Store Metrics
+      self.EvaluationList[f"Eval_{FitName}_{ScoredDataName}"] = Metrics
